/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "ginfo.h"

void modules_xml_null (void *, const char *, ...);
char *modules_xml_server_name (void);
size_t modules_xml_server_write (void *, size_t, size_t, void *);
void modules_xml_tmp_dir_mkdir (char *);

int
modules_xml_check (char *a)
{
  xmlDocPtr doc;
  xmlNodePtr cur;

  if (!(doc = xmlParseFile (a)))
    {
      msg ("Module %s: syntax error.", a);
      return 1;
    }

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      msg ("Module %s: syntax error.", a);
      xmlFreeDoc (doc);
      return 1;
    }

  if (!xmlStrcmp (cur->name, "article"))
    {
      if (modules_docbook_check (doc, cur, a))
	{
	  xmlFreeDoc (doc);
	  return 1;
	}
      else
	return 0;
    }

  else if (!xmlStrcmp (cur->name, "ginfo"))
    {
      if (modules_ginfo_check (doc, cur, a))
	{
	  xmlFreeDoc (doc);
	  return 1;
	}
      else
	return 0;
    }

  else
    {
      msg ("Module %s: syntax error.", a);
      xmlFreeDoc (doc);
      return 1;
    }
}

/* Download remote module functions: */
char *
modules_xml_server_name (void)
{
  char *f;
  int d = 0;
  int len =
    strlen (modules_xml_tmp_dir (GINFO_TMP_DIR)) +
    strlen ("/ginfo_module_.xml") + 4;
  struct stat st;

  if (!(f = (char *) malloc (sizeof (char) * len)))
    fatal ("Error: memory.");

  while (d < 1000)
    {
      snprintf (f, len, "%s/ginfo_module_%d.xml",
		modules_xml_tmp_dir (GINFO_TMP_DIR), d);

      if (lstat (f, &st))
	break;

      d++;
    }

  if (d == 1000)
    {
      free (f);
      return NULL;
    }

  return f;
}

size_t
modules_xml_server_write (void *ptr, size_t size, size_t nmemb, void *data)
{
  register int realsize = size * nmemb;
  FILE *fl = (FILE *) data;

  fwrite (ptr, realsize, 1, fl);

  return realsize;
}

void
modules_xml_null (void *fl, const char *dummy, ...)
{
  /* Be happy :) */
}

int
modules_xml_server (char *server)
{
  char *file;

  /* This is an hack :/ */
  if (!mainwin)
    {
      modules_graphic_text ("Download %s", server);

      if (!nobootsplash)
	{
	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	}
    }

  if (!(file = modules_xml_download (server)))
    return 1;

  modules_xml_server_check_list (file,
				 modules_xml_server_check_name (server));

  return 0;
}

char *
modules_xml_server_check_name (char *server_remote)
{
  int len;
  char *server;

  if (!strncmp ("http://", server_remote, 7))
    server = server_remote + 7;
  else
    server = server_remote;

  len = strlen (server);

  while (len)
    {
      if (server[len] == '/')
	break;
      len--;
    }

  if (!len)
    return server_remote;

  server[len] = 0;
  return server_remote;
}

void
modules_xml_server_check_list (char *file, char *server)
{
  xmlDocPtr doc;
  FILE *fl;
  register int i = 0;
  char buf[1024];
  char *tmp;
  int len;

  xmlSetGenericErrorFunc (xmlGenericErrorContext, modules_xml_null);

  if ((doc = xmlParseFile (file)))
    {
      xmlFree (doc);
      return;
    }

  if (!(fl = fopen (file, "r")))
    return;

  while ((buf[i] = (char) getc (fl)) != EOF)
    {
      if (buf[i] == '\n')
	{
	  buf[i] = 0;

	  if (strncmp (buf, "http://", 7))
	    {

	      len = strlen (buf) + strlen (server) + 2;
	      if (!(tmp = (char *) malloc (sizeof (char) * len)))
		fatal ("Error: memory.");

	      snprintf (tmp, len, "%s/%s", server, buf);
	      modules_xml_server (tmp);

	      free (tmp);

	    }
	  else
	    modules_xml_server (buf);

	  i = 0;
	}
      else if (i < 1024)
	i++;
    }

  fclose (fl);
  unlink (file);
}

char *
modules_xml_download (char *server)
{
  CURL *curl;
  FILE *fl;
  char *file = modules_xml_server_name ();
  long code;

  if (!file)
    return NULL;

  if (!(fl = fopen (file, "w")))
    {
      free (file);
      return NULL;
    }

  curl_global_init (CURL_GLOBAL_DEFAULT);

  if (!(curl = curl_easy_init ()))
    {
      free (file);
      return NULL;
    }

  curl_easy_setopt (curl, CURLOPT_URL, server);
  curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, modules_xml_server_write);
  curl_easy_setopt (curl, CURLOPT_FILE, (void *) fl);
  curl_easy_setopt (curl, CURLOPT_TIMEOUT, 10);

  if (curl_easy_perform (curl))
    {
      curl_easy_cleanup (curl);
      curl_global_cleanup ();

      free (file);
      return NULL;
    }

  curl_easy_getinfo (curl, CURLINFO_HTTP_CODE, &code);
  if (code != 200)
    {
      curl_easy_cleanup (curl);
      curl_global_cleanup ();

      fclose (fl);
      unlink (file);

      free (file);

      return NULL;
    }

  curl_easy_cleanup (curl);
  curl_global_cleanup ();

  fclose (fl);

  return file;
}

void
modules_xml_delete (void)
{
  DIR *d;
  struct dirent *dir;
  char *tmp;

  if (!(d = opendir (modules_xml_tmp_dir (GINFO_TMP_DIR))))
    return;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strncmp (dir->d_name, "ginfo_module_", strlen ("ginfo_module_")))
	continue;

      n =
	strlen (dir->d_name) + strlen (modules_xml_tmp_dir (GINFO_TMP_DIR)) +
	2;
      if (!(tmp = (char *) malloc (sizeof (char) * n)))
	fatal ("Error: memory.");

      snprintf (tmp, n, "%s/%s", modules_xml_tmp_dir (GINFO_TMP_DIR),
		dir->d_name);
      unlink (tmp);

      free (tmp);
    }
}


/* tmp_dir */
char *
modules_xml_tmp_dir (char *a)
{
  static char *ret = NULL;
  int len;
  char *home;
  char *tmp;

  if (ret)
    {
      modules_xml_tmp_dir_mkdir (ret);
      return ret;
    }

  if (!(home = getenv ("HOME")))
    {
      ret = a;
      modules_xml_tmp_dir_mkdir (ret);
      return ret;
    }

  len = strlen (home) + strlen ("/.ginfo") + 1;
  if (!(tmp = (char *) malloc (sizeof (char) * len)))
    fatal ("Error: memory.");

  snprintf (tmp, len, "%s/.ginfo", home);

  ret = tmp;

  modules_xml_tmp_dir_mkdir (ret);

  return ret;
}

void
modules_xml_tmp_dir_mkdir (char *f)
{
  struct stat st;

  if (lstat (f, &st))
    mkdir (f, 0750);
}

char *
modules_xml_info_trim (char *a)
{
  register int i = 0, j = 0;
  char *b;
  int k = 0, len;

  if (!a || !*a)
    return NULL;

  len = strlen (a);

  if (!(b = (char *) malloc (sizeof (char) * (len + 1))))
    fatal ("Error: memory.");

  while (i < len)
    {
      if (a[i] == ' ' || a[i] == '\t' || a[i] == '\r' || a[i] == '\n')
	{
	  if (!k)
	    {
	      b[j++] = ' ';
	      k = 1;
	    }
	}
      else
	{
	  b[j++] = a[i];
	  k = 0;
	}

      i++;
    }

  b[j] = 0;

  for (i = k = 0; i < j; i++)
    if (b[i] != ' ')
      k++;

  if (!k)
    {
      free (b);
      return NULL;
    }

  return b;
}

/* EOF */
