/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include <xml2c.h>

struct output_data
{
  char *dir;
  char *output;
  char *no_ext;
};

static struct output_data output_file;

void check_output_data (char *);
void create_package ();
void error (void *, const char *, ...);
char *trim (char *);
ginfo_tips *check_tips (xmlDocPtr, xmlNodePtr, int *);
ginfo_docs *check_docs (xmlDocPtr, xmlNodePtr, int *);
char *check_genre (char *);
xml_info *check_info (xmlDocPtr, xmlNodePtr);
void check_parser_rec (xmlDocPtr, xmlNodePtr, xml_info **,
		       xml_info **, xml_font *);
char *trim_info (char *);
int check_tag (xmlDocPtr, xmlNodePtr, char *, xml_font *);
void write_output (ginfo_docs *, int, ginfo_tips *, int, int, char *, char *,
		   char **);

int
main (int argc, char *argv[])
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  char *def = NULL;
  char *name = NULL;
  char *genre = NULL;
  char *words = NULL;
  char **words_list = NULL;
  ginfo_tips *tips = NULL;
  ginfo_docs *docs = NULL;
  int docs_n = 0;
  int tips_n = 0;
  char *genre_id = NULL;
  char *input = NULL;
  int i, package = 0;

  struct stat st;

  if (argc < 2)
    usage (1);

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-h") || !strcmp (argv[1], "--help"))
	usage (0);

      else if (!strcmp (argv[i], "-p") || !strcmp (argv[1], "--package"))
	package = 1;

      else if (!input)
	input = argv[i];

      else if (!output_file.dir)
	check_output_data (argv[i]);

      else
	usage (1);
    }

  if (!input)
    usage (1);

  if (!output_file.dir)
    check_output_data (input);

  if (package)
    {
      char *o;

      if (!
	  (o =
	   (char *) malloc (sizeof (char) *
			    (strlen (output_file.no_ext) +
			     strlen (output_file.dir) + 2))))
	fatal ("Error: memory.");

      sprintf (o, "%s/%s", output_file.dir, output_file.no_ext);

      output_file.dir = o;

      if (mkdir (output_file.dir, 0750) && errno != EEXIST)
	fatal ("Error: mkdir %s", output_file.dir);

      create_package ();
    }

  xmlSetGenericErrorFunc (xmlGenericErrorContext, error);

  if (!(doc = xmlParseFile (input)))
    fatal ("XML syntax error.");

  if (lstat (input, &st))
    fatal ("I can't open %s!", input);

  xmlGenericError (xmlGenericErrorContext, "Parser ok!\n");

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      xmlFreeDoc (doc);
      fatal ("No root element 'ginfo'");
    }

  if (xmlStrcmp (cur->name, "ginfo"))
    {
      xmlGenericError (xmlGenericErrorContext,
		       "RootElement must by \"ginfo\"!\n");
      xmlFreeDoc (doc);
      fatal ("No root element 'ginfo'");
    }

  xmlGenericError (xmlGenericErrorContext, "RootElement ok!\n");

  for (cur = cur->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "module"))
	{
	  def = xmlGetProp (cur, "default");

	  if (!(name = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*name)
	    {
	      xmlFreeDoc (doc);
	      fatal ("No name set!");
	    }

	  if (def)
	    def = trim (def);
	  name = trim (name);

	}

      else if (!xmlStrcmp (cur->name, "genre"))
	{
	  if (!(genre = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*genre)
	    {
	      xmlFreeDoc (doc);
	      fatal ("No genre set!");
	    }

	  genre = trim (genre);
	}

      else if (!xmlStrcmp (cur->name, "words"))
	{
	  if (!(words = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*words)
	    {
	      xmlFreeDoc (doc);
	      fatal ("No words set!");
	    }

	  words = trim (words);
	}

      else if (!xmlStrcmp (cur->name, "tooltips"))
	{
	  if (!(tips = check_tips (doc, cur, &tips_n)))
	    {
	      xmlFreeDoc (doc);
	      fatal ("No tip element!");
	    }
	}

      else if (!xmlStrcmp (cur->name, "docs"))
	{
	  if (!(docs = check_docs (doc, cur, &docs_n)))
	    {
	      xmlFreeDoc (doc);
	      fatal ("No doc element!");
	    }
	}
    }

  if (!name)
    {
      xmlFreeDoc (doc);
      fatal ("No name set!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module name: %s\n", name);

  if (def)
    xmlGenericError (xmlGenericErrorContext, "Set flag default: YES\n");
  else
    xmlGenericError (xmlGenericErrorContext, "Set flag default: NO\n");

  if (!genre)
    {
      xmlFreeDoc (doc);
      fatal ("No genre set!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module genre: %s\n", genre);

  if (genre && !(genre_id = check_genre (genre)))
    {
      xmlFreeDoc (doc);
      fatal ("No compatible genre!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module genre Id: %d\n", genre_id);

  if (!words)
    {
      xmlFreeDoc (doc);
      fatal ("No words set!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module words: %s\n", words);

  if (!(words_list = ginfo_check_words (words)))
    {
      xmlFreeDoc (doc);
      fatal ("No compatible words!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module words list: ok!\n");

  if (!docs)
    {
      xmlFreeDoc (doc);
      fatal ("No docs!");
    }

  xmlGenericError (xmlGenericErrorContext, "Module docs list: ok!\n");

  if (tips)
    xmlGenericError (xmlGenericErrorContext, "Module tips list: ok!\n");
  else
    xmlGenericError (xmlGenericErrorContext, "Module tips list: no.\n");

  write_output (docs, docs_n, tips, tips_n, atoi (def), name, genre_id,
		words_list);

  return 0;
}

/* Write in parser error */
void
error (void *fl, const char *str, ...)
{
  va_list va;

  va_start (va, str);

  vfprintf (stderr, str, va);

  va_end (va);
}

/* Trim the tag */
char *
trim (char *tmp)
{
  int i = 0;
  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    tmp++;

  i = strlen (tmp);
  i--;

  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    i--;

  tmp[i + 1] = 0;

  return tmp;
}

/* Check tips tags */
ginfo_tips *
check_tips (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur;
  char *tip;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "tip"))
	{
	  if (!(tip = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*tip)
	    {
	      xmlGenericError (xmlGenericErrorContext, "no Tip set!\n");
	      return NULL;
	    }

	  tip = trim (tip);

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  ginfo_easy_tips_add (t, tip, NULL, NULL, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_tips_close (t, n);
}

/* Check the docs tags */
ginfo_docs *
check_docs (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur, cur_c;
  char *chapter;
  xml_info *info;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "doc"))
	{
	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  chapter = NULL;
	  info = NULL;
	  for (cur_c = cur->children; cur_c != NULL; cur_c = cur_c->next)
	    {

	      if (!xmlStrcmp (cur_c->name, "chapter"))
		{
		  if (!
		      (chapter =
		       xmlNodeListGetString (doc, cur_c->xmlChildrenNode, 1))
		      || !*chapter)
		    {
		      ginfo_easy_docs_free (t);

		      xmlGenericError (xmlGenericErrorContext,
				       "no chapter set!\n");

		      return NULL;
		    }

		  chapter = trim (chapter);

		}
	      else if (!xmlStrcmp (cur_c->name, "info"))
		{
		  if (!(info = check_info (doc, cur_c->children)))
		    {
		      xmlGenericError (xmlGenericErrorContext,
				       "no info text set!\n");
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }
		}
	    }
	  if (!chapter)
	    {
	      xmlGenericError (xmlGenericErrorContext, "no chapter set!\n");
	      ginfo_easy_docs_free (t);
	      return NULL;
	    }

	  ginfo_easy_docs_add (t, chapter, NULL, (void *) info, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_docs_close (t, n);
}

/* check info tags. This functions run only check_parser_rec.
 * the paeser_check_info is a recursive function.
 */
xml_info *
check_info (xmlDocPtr doc, xmlNodePtr cur)
{
  xml_info *t = NULL;
  xml_info *h = NULL;
  xml_font font;

  memset (&font, 0x0, sizeof (font));

  check_parser_rec (doc, cur, &t, &h, &font);

  return h;
}

/* The recursive function for info tags */
void
check_parser_rec (xmlDocPtr doc, xmlNodePtr cur, xml_info ** t,
		  xml_info ** h, xml_font * font)
{

  xml_info *c;
  char *link = NULL;

  if (*t)
    link = (*t)->link;

  while (cur)
    {

      c = NULL;

      if (cur->type == XML_TEXT_NODE)
	{
	  if (!(c = (xml_info *) malloc (sizeof (xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->image = NULL;
	  c->link = link;

	  c->type = INFO_TEXT;
	  c->text = xmlNodeGetContent (cur);
	  if (c->text)
	    c->text = trim_info (c->text);

	  if (!c->text)
	    {
	      free (c);
	      c = NULL;
	    }
	}
      else if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!(c = (xml_info *) malloc (sizeof (xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->text = NULL;
	  c->image = NULL;
	  c->link = link;

	  if (!strcmp (cur->name, "nl"))
	    c->type = INFO_NEWLINE;

	  else if (!strcmp (cur->name, "space"))
	    c->type = INFO_SPACE;

	  else if (!strcmp (cur->name, "image"))
	    {
	      c->type = INFO_IMAGE;

	      if (!
		  (c->image =
		   xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
		  || !*c->image)
		{
		  free (c);
		  c = NULL;
		}
	    }

	  else
	    {
	      c->type = INFO_FONT;
	      c->link = link;

	      memcpy (&c->font, font, sizeof (xml_font));

	      if (!strcmp (cur->name, "link"))
		{
		  c->link = xmlGetProp (cur, "id");
		  if (!c->link || !*c->link)
		    {
		      free (c);
		      c = NULL;
		    }
		}

	      else if (check_tag (doc, cur, (char *) cur->name, &c->font))
		{
		  free (c);
		  c = NULL;
		}
	    }
	}

      if (c)
	{
	  if (!*h)
	    *h = c;
	  else
	    (*t)->next = c;

	  *t = c;

	  if (c->type == INFO_FONT)
	    {
	      check_parser_rec (doc, cur->children, t, h, &c->font);
	      if (!(c = (xml_info *) malloc (sizeof (xml_info))))
		fatal ("Error: memory.");

	      c->next = NULL;
	      c->text = NULL;
	      c->link = link;
	      (*t)->next = c;
	      *t = c;

	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (xml_font));
	    }
	}

      cur = cur->next;
    }
}

/* Search the correct tags */
int
check_tag (xmlDocPtr doc, xmlNodePtr cur, char *tag, xml_font * font)
{
  /* STYLE */
  if (!strcmp (tag, "oblique"))
    {
      font->style = "PANGO_STYLE_OBLIQUE";
      return 0;
    }

  if (!strcmp (tag, "italic"))
    {
      font->style = "PANGO_STYLE_ITALIC";
      return 0;
    }

  /* VARIANT */
  if (!strcmp (tag, "small_caps"))
    {
      font->variant = "PANGO_VARIANT_SMALL_CAPS";
      return 0;
    }

  /* WEIGHT */
  if (!strcmp (tag, "ultralight"))
    {
      font->weight = "PANGO_WEIGHT_ULTRALIGHT";
      return 0;
    }

  if (!strcmp (tag, "light"))
    {
      font->weight = "PANGO_WEIGHT_LIGHT";
      return 0;
    }

  if (!strcmp (tag, "bold"))
    {
      font->weight = "PANGO_WEIGHT_BOLD";
      return 0;
    }

  if (!strcmp (tag, "ultrabold"))
    {
      font->weight = "PANGO_WEIGHT_ULTRABOLD";
      return 0;
    }

  if (!strcmp (tag, "heavy"))
    {
      font->weight = "PANGO_WEIGHT_HEAVY";
      return 0;
    }

  /* STRETCH */
  if (!strcmp (tag, "ultracondensed"))
    {
      font->stretch = "PANGO_STRETCH_ULTRA_CONDENSED";
      return 0;
    }

  if (!strcmp (tag, "extracondensed"))
    {
      font->stretch = "PANGO_STRETCH_EXTRA_CONDENSED";
      return 0;
    }

  if (!strcmp (tag, "condensed"))
    {
      font->stretch = "PANGO_STRETCH_CONDENSED";
      return 0;
    }

  if (!strcmp (tag, "semicondensed"))
    {
      font->stretch = "PANGO_STRETCH_SEMI_CONDENSED";
      return 0;
    }

  if (!strcmp (tag, "semiexpanded"))
    {
      font->stretch = "PANGO_STRETCH_SEMI_EXPANDED";
      return 0;
    }

  if (!strcmp (tag, "expanded"))
    {
      font->stretch = "PANGO_STRETCH_EXPANDED";
      return 0;
    }

  if (!strcmp (tag, "extraexpanded"))
    {
      font->stretch = "PANGO_STRETCH_EXTRA_EXPANDED";
      return 0;
    }

  if (!strcmp (tag, "ultraexpanded"))
    {
      font->stretch = "PANGO_STRETCH_ULTRA_EXPANDED";
      return 0;
    }

  /* SCALE */
  if (!strcmp (tag, "xx_small"))
    {
      font->scale = "PANGO_SCALE_XX_SMALL";
      return 0;
    }

  if (!strcmp (tag, "x_small"))
    {
      font->scale = "PANGO_SCALE_X_SMALL";
      return 0;
    }

  if (!strcmp (tag, "small"))
    {
      font->scale = "PANGO_SCALE_SMALL";
      return 0;
    }

  if (!strcmp (tag, "large"))
    {
      font->scale = "PANGO_SCALE_LARGE";
      return 0;
    }

  if (!strcmp (tag, "x_large"))
    {
      font->scale = "PANGO_SCALE_X_LARGE";
      return 0;
    }

  if (!strcmp (tag, "xx_large"))
    {
      font->scale = "PANGO_SCALE_XX_LARGE";
      return 0;
    }

  /* SIZE */
  if (!strcmp (tag, "size"))
    {
      char *size;
      size = xmlGetProp (cur, "value");
      if (!size || !*size)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      font->size = size;

      return 0;
    }

  /* COLOR */
  if (!strcmp (tag, "foreground"))
    {
      font->foreground = xmlGetProp (cur, "value");
      if (!font->foreground || !*font->foreground)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "background"))
    {
      font->background = xmlGetProp (cur, "value");
      if (!font->background || !*font->background)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  /* JUSTIFY */
  if (!strcmp (tag, "justify"))
    {
      char *justify;
      justify = xmlGetProp (cur, "value");

      if (!justify || !*justify)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      if (!strcmp (justify, "center"))
	font->justify = "GTK_JUSTIFY_CENTER";
      else if (!strcmp (justify, "left"))
	font->justify = "GTK_JUSTIFY_LEFT";
      else if (!strcmp (justify, "right"))
	font->justify = "GTK_JUSTIFY_RIGHT";
      else if (!strcmp (justify, "fill"))
	font->justify = "GTK_JUSTIFY_FILL";
      else
	{
	  xmlGenericError (xmlGenericErrorContext,
			   "tag %s value=center|left|right\n", tag);
	  return 1;
	}

      return 0;
    }

  /* UNDERLINE */
  if (!strcmp (tag, "underline"))
    {
      font->underline = 1;
      return 0;
    }

  if (!strcmp (tag, "double_underline"))
    {
      font->underline = 2;
      return 0;
    }

  /* FACE */
  if (!strcmp (tag, "face"))
    {
      font->face = xmlGetProp (cur, "value");
      if (!font->face || !*font->face)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  /* MARGIN */
  if (!strcmp (tag, "margin"))
    {
      char *v;

      v = xmlGetProp (cur, "left");
      if (v)
	{
	  font->margin_left = atoi (v);
	  if (font->margin_left < 0)
	    font->margin_left = 0;
	}

      v = xmlGetProp (cur, "right");
      if (v)
	{
	  font->margin_right = atoi (v);
	  if (font->margin_right < 0)
	    font->margin_right = 0;
	}

      return 0;
    }

  /* FONT */
  if (!strcmp (tag, "font"))
    {
      char *tmp;
      tmp = xmlGetProp (cur, "style");

      /* Style */
      if (tmp)
	{
	  if (!strcmp (tmp, "oblique"))
	    font->style = "PANGO_STYLE_OBLIQUE";
	  else if (!strcmp (tmp, "italic"))
	    font->style = "PANGO_STYLE_ITALIC";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Style axpected oblique|italic");
	      return 1;
	    }
	}

      /* Variant */
      tmp = xmlGetProp (cur, "variant");
      if (tmp)
	{
	  if (!strcmp (tmp, "small_cups"))
	    font->variant = "PANGO_VARIANT_SMALL_CAPS";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Variant axpected small_caps");
	      return 1;
	    }
	}

      /* Weight */
      tmp = xmlGetProp (cur, "weight");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultralight"))
	    font->weight = "PANGO_WEIGHT_ULTRALIGHT";
	  else if (!strcmp (tmp, "light"))
	    font->weight = "PANGO_WEIGHT_LIGHT";
	  else if (!strcmp (tmp, "bold"))
	    font->weight = "PANGO_WEIGHT_BOLD";
	  else if (!strcmp (tmp, "ultrabold"))
	    font->weight = "PANGO_WEIGHT_ULTRABOLD";
	  else if (!strcmp (tmp, "heavy"))
	    font->weight = "PANGO_WEIGHT_HEAVY";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Weight axpected ultralight|light|bold|ultrabold|heavy");
	      return 1;
	    }
	}

      /* Stretch */
      tmp = xmlGetProp (cur, "stretch");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultracondensed"))
	    font->stretch = "PANGO_STRETCH_ULTRA_CONDENSED";
	  else if (!strcmp (tmp, "extracondensed"))
	    font->stretch = "PANGO_STRETCH_EXTRA_CONDENSED";
	  else if (!strcmp (tmp, "condensed"))
	    font->stretch = "PANGO_STRETCH_CONDENSED";
	  else if (!strcmp (tmp, "semicondensed"))
	    font->stretch = "PANGO_STRETCH_SEMI_CONDENSED";
	  else if (!strcmp (tmp, "semiexpanded"))
	    font->stretch = "PANGO_STRETCH_SEMI_EXPANDED";
	  else if (!strcmp (tmp, "expanded"))
	    font->stretch = "PANGO_STRETCH_EXPANDED";
	  else if (!strcmp (tmp, "extraexpanded"))
	    font->stretch = "PANGO_STRETCH_EXTRA_EXPANDED";
	  else if (!strcmp (tmp, "ultraexpanded"))
	    font->stretch = "PANGO_STRETCH_ULTRA_EXPANDED";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Stretch axpected ultracondensed|extracondensed|condensed|semicondensed|semiexpanded|expanded|extraexpanded|ultraexpanded");
	      return 1;
	    }
	}

      /* Scale */
      tmp = xmlGetProp (cur, "scale");
      if (tmp)
	{
	  if (!strcmp (tmp, "xx_small"))
	    font->scale = "PANGO_SCALE_XX_SMALL";
	  if (!strcmp (tmp, "x_small"))
	    font->scale = "PANGO_SCALE_X_SMALL";
	  if (!strcmp (tmp, "small"))
	    font->scale = "PANGO_SCALE_SMALL";
	  if (!strcmp (tmp, "large"))
	    font->scale = "PANGO_SCALE_LARGE";
	  if (!strcmp (tmp, "x_large"))
	    font->scale = "PANGO_SCALE_X_LARGE";
	  if (!strcmp (tmp, "xx_large"))
	    font->scale = "PANGO_SCALE_XX_LARGE";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Scale axpected xx_small|x_small|small|large|x_lage|xx_large");
	      return 1;
	    }
	}

      /* Size */
      tmp = xmlGetProp (cur, "size");
      if (tmp)
	{
	  font->size = tmp;
	}

      /* Color */
      font->foreground = xmlGetProp (cur, "foreground");
      font->background = xmlGetProp (cur, "background");

      /* Justify */
      tmp = xmlGetProp (cur, "justify");
      if (tmp)
	{
	  if (!strcmp (tmp, "center"))
	    font->justify = "GTK_JUSTIFY_CENTER";
	  else if (!strcmp (tmp, "left"))
	    font->justify = "GTK_JUSTIFY_LEFT";
	  else if (!strcmp (tmp, "right"))
	    font->justify = "GTK_JUSTIFY_RIGHT";
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Justify axpected center|left|right");
	      return 1;
	    }
	}

      /* Underline */
      tmp = xmlGetProp (cur, "underline");
      if (tmp)
	{
	  if (!strcmp (tmp, "single"))
	    font->underline = 1;
	  else if (!strcmp (tmp, "double"))
	    font->underline = 2;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Underline axpected single|double");
	      return 1;
	    }
	}

      /* Face */
      font->face = xmlGetProp (cur, "face");

      /* margin_left */
      tmp = xmlGetProp (cur, "margin_left");
      if (tmp)
	{
	  font->margin_left = atoi (tmp);
	  if (font->margin_left < 0)
	    font->margin_left = 0;
	}

      /* margin_right */
      tmp = xmlGetProp (cur, "margin_right");
      if (tmp)
	{
	  font->margin_right = atoi (tmp);
	  if (font->margin_right < 0)
	    font->margin_right = 0;
	}

      return 0;
    }

  xmlGenericError (xmlGenericErrorContext, "Unknow tag %s!\n", tag);

  return 1;
}

/* trim the info text */
char *
trim_info (char *a)
{
  register int i = 0, j = 0;
  char *b;
  int k = 0, len;

  if (!a || !*a)
    return NULL;

  len = strlen (a);

  if (!(b = (char *) malloc (sizeof (char) * (len + 1))))
    fatal ("Error: memory.");

  while (i < len)
    {
      if (a[i] == ' ' || a[i] == '\t' || a[i] == '\r' || a[i] == '\n')
	{
	  if (!k)
	    {
	      b[j++] = ' ';
	      k = 1;
	    }
	}
      else
	{
	  b[j++] = a[i];
	  k = 0;
	}

      i++;
    }

  b[j] = 0;

  for (i = k = 0; i < j; i++)
    if (b[i] != ' ')
      k++;

  if (!k)
    {
      free (b);
      return NULL;
    }

  return b;
}

char *
check_genre (char *g)
{

  if (!strcasecmp (g, "SYSTEM"))
    return "GINFO_GENRE_SYSTEM";
  if (!strcasecmp (g, "NETWORK"))
    return "GINFO_GENRE_NETWORK";
  if (!strcasecmp (g, "AUDIO"))
    return "GINFO_GENRE_AUDIO";
  if (!strcasecmp (g, "VIDEO"))
    return "GINFO_GENRE_VIDEO";
  if (!strcasecmp (g, "STREAM"))
    return "GINFO_GENRE_STREAM";
  if (!strcasecmp (g, "OFFICE"))
    return "GINFO_GENRE_OFFICE";
  if (!strcasecmp (g, "CODING"))
    return "GINFO_GENRE_CODING";

  return NULL;
}

void
fatal (char *str, ...)
{
  va_list va;

  va_start (va, str);

  fprintf (stderr, "* ");
  vfprintf (stderr, str, va);
  fprintf (stderr, "\n");

  va_end (va);

  exit (1);
}

void
usage (int r)
{

  fprintf (stderr, "%s %s - %s - %s\n\n", NAME, VERSION, AUTHOR, COPYRIGHT);
  fprintf (stderr,
	   "\t%s [-h|--help] [-p|--package] input.xml [output.c|directory]\n\n",
	   NAME);

  exit (r);
}

/* Write C file */
void
write_output (ginfo_docs * docs, int docs_n, ginfo_tips * tips, int tips_n,
	      int def, char *name, char *genre, char **words)
{
  FILE *fl;
  char *o;
  int i, j;
  xml_info *info;

  if (!
      (o =
       (char *) malloc (sizeof (char) *
			(strlen (output_file.output) +
			 strlen (output_file.dir) + 2))))
    fatal ("Error: memory.");

  sprintf (o, "%s/%s", output_file.dir, output_file.output);

  if (!(fl = fopen (o, "w")))
    fatal ("I can't open %s.", o);
  fprintf (stderr, "\nWrite %s.\n", o);

  /* INIT */
  fprintf (fl,
	   "/* This source is generated by %s */\n\n"
	   "#include <ginfo/ginfo_plugin.h>\n#include <stdio.h>\n"
	   "#include <string.h>\n#include <stdlib.h>\n\n"
	   "typedef enum\n{\n" "  INFO_TEXT,\n" "  INFO_NEWLINE,\n"
	   "  INFO_SPACE,\n" "  INFO_IMAGE,\n" "  INFO_FONT\n"
	   "} module_info_tag;\n\n" "typedef struct\n{\n"
	   "  PangoStyle style;\n" "  PangoVariant variant;\n"
	   "  PangoWeight weight;\n" "  PangoStretch stretch;\n"
	   "  double scale;\n" "  int size;\n" "  char *foreground;\n"
	   "  char *background;\n" "  GtkJustification justify;\n"
	   "  int underline;\n" "  char *face;\n" "  int margin_left;\n"
	   "  int margin_right;\n" "} module_info_font;\n\n"
	   "typedef struct _module_info_\n{\n" "  module_info_tag type;\n"
	   "  char *text;\n" "  char *image;\n" "  char *link;\n"
	   "  module_info_font font;\n" "  struct _module_info_ *next;\n"
	   "} module_info;\n\n", NAME);

  fprintf (stderr, "Init: done\n");

  /* DEFAULT */
  fprintf (fl, "int\nmodule_default (void)\n{\n  return %d;\n}\n\n", def);
  fprintf (stderr, "Default: done\n");

  /* NAME */
  fprintf (fl, "char *\nmodule_name (void)\n{\n  return \"%s\";\n}\n\n",
	   name);
  fprintf (stderr, "Name: done\n");

  /* GENRE */
  fprintf (fl, "int\nmodule_genre (void)\n{\n  return %s;\n}\n\n", genre);
  fprintf (stderr, "Genre: done\n");

  /* WORDS */
  fprintf (fl, "char **\nmodule_words (void)\n{\n");
  fprintf (stderr, "Words: ");
  i = 0;
  while (words[i])
    i++;
  fprintf (fl,
	   "  char **list;\n\n  if (!(list = (char **) malloc (sizeof (char *) * %d)))\n    ginfo_error (\"Error: memory.\");\n\n",
	   i + 1);

  for (j = 0; j < i; j++)
    {
      fprintf (fl, "  list[%d] = \"%s\";\n", j, words[j]);
      fprintf (stderr, ".");
    }
  fprintf (fl, "  list[%d] = NULL;\n", i);
  fprintf (fl, "\n  return list;\n}\n\n");
  fprintf (stderr, " done.\n");

  /* TIPS */
  if (tips)
    {
      fprintf (stderr, "Tips: ");
      fprintf (fl,
	       "ginfo_tips *\nmodule_tooltips (int *n)\n{\n  void *t;\n\n  t = ginfo_easy_tips_init ();\n\n");

      for (i = 0; i < tips_n; i++)
	{
	  fprintf (fl, "  ginfo_easy_tips_add (t, \"%s\", NULL, NULL, ",
		   tips[i].text);
	  if (!tips[i].id)
	    fprintf (fl, "NULL);\n");
	  else
	    fprintf (fl, "\"%s\");\n", tips[i].id);
	  fprintf (stderr, ".");
	}

      fprintf (fl, "\n  return ginfo_easy_tips_close (t, n);\n}\n\n");
      fprintf (stderr, " done.\n");
    }

  /* DOCS */
  fprintf (stderr, "Docs: ");
  fprintf (fl,
	   "ginfo_docs *\nmodule_docs (int *n)\n{\n  void *t;\n  module_info *info, *info_new;\n\n  t = ginfo_easy_docs_init ();\n\n  info = info_new = NULL;\n\n");

  for (j = i = 0; i < docs_n; i++)
    {
      info = (xml_info *) docs[i].arg;

      while (info)
	{
	  fprintf (fl,
		   "  info_new = (module_info *) malloc (sizeof (module_info));\n");
	  fprintf (fl,
		   "  if (!info_new)\n    ginfo_error (\"Error: memory.\");\n");

	  if (info->link)
	    fprintf (fl, "  info_new->link = \"%s\";\n", info->link);
	  else
	    fprintf (fl, "  info_new->link = NULL;\n");

	  switch (info->type)
	    {
	    case INFO_TEXT:
	      fprintf (fl, "  info_new->type = INFO_TEXT;\n");
	      fprintf (fl, "  info_new->text = \"%s\";\n", info->text);
	      break;
	    case INFO_NEWLINE:
	      fprintf (fl, "  info_new->type = INFO_NEWLINE;\n");
	      break;
	    case INFO_SPACE:
	      fprintf (fl, "  info_new->type = INFO_SPACE;\n");
	      break;
	    case INFO_IMAGE:
	      fprintf (fl, "  info_new->type = INFO_IMAGE;\n");
	      fprintf (fl, "  info_new->image = \"%s\";\n", info->image);
	      break;
	    case INFO_FONT:
	      fprintf (fl, "  info_new->type = INFO_FONT;\n");

	      fprintf (fl,
		       "  memset (&info_new->font, 0, sizeof (module_info_font));\n");
	      if (info->font.style)
		fprintf (fl, "  info_new->font.style = %s;\n",
			 info->font.style);
	      if (info->font.variant)
		fprintf (fl, "  info_new->font.variant = %s;\n",
			 info->font.variant);
	      if (info->font.weight)
		fprintf (fl, "  info_new->font.weight = %s;\n",
			 info->font.weight);
	      if (info->font.stretch)
		fprintf (fl, "  info_new->font.stretch = %s;\n",
			 info->font.stretch);
	      if (info->font.scale)
		fprintf (fl, "  info_new->font.scale = %s;\n",
			 info->font.scale);
	      else
		fprintf (fl,
			 "  info_new->font.scale = PANGO_SCALE_MEDIUM;\n");
	      if (info->font.size)
		fprintf (fl, "  info_new->font.size = %s * PANGO_SCALE;\n",
			 info->font.size);
	      if (info->font.foreground)
		fprintf (fl, "  info_new->font.foreground = \"%s\";\n",
			 info->font.foreground);
	      if (info->font.background)
		fprintf (fl, "  info_new->font.background = \"%s\";\n",
			 info->font.background);
	      if (info->font.justify)
		fprintf (fl, "  info_new->font.justify = %s;\n",
			 info->font.justify);
	      if (info->font.underline)
		fprintf (fl, "  info_new->font.underline = %d;\n",
			 info->font.underline);
	      if (info->font.face)
		fprintf (fl, "  info_new->font.face = \"%s\";\n",
			 info->font.face);
	      if (info->font.margin_left)
		fprintf (fl, "  info_new->font.margin_left = %d;\n",
			 info->font.margin_left);
	      if (info->font.margin_right)
		fprintf (fl, "  info_new->font.margin_right = %d;\n",
			 info->font.margin_right);
	      break;
	    }

	  fprintf (fl, "\n");

	  if (!j)
	    {
	      fprintf (fl,
		       "  ginfo_easy_docs_add (t, \"%s\", ginfo_tutor_xml, (void *) info_new, ",
		       docs[i].text);
	      if (!docs[i].id)
		fprintf (fl, "NULL);\n");
	      else
		fprintf (fl, "\"%s\");\n", docs[i].id);

	      fprintf (fl, "  info = info_new;\n\n");

	      j++;
	    }
	  else if (info->next)
	    {
	      fprintf (fl, "  info->next = info_new;\n");
	      fprintf (fl, "  info = info_new;\n\n");
	    }
	  else
	    fprintf (fl, "  info_new->next = NULL;\n\n");

	  info = info->next;
	}

      fprintf (stderr, ".");
    }

  fprintf (fl, "  return ginfo_easy_docs_close (t, n);\n}\n\n");
  fprintf (stderr, " done.\n");

  fprintf (stderr, "Close %s.\nbye!\n", o);
  fclose (fl);
  free (o);
}

char *
output_no_extension (char *o)
{
  static char output[1024];
  register int i = 0, len;

  strncpy (output, o, 1024);
  len = strlen (output) - 1;

  while (i < len)
    {
      if (!((output[i] >= 'A' && output[i] <= 'Z')
	    || (output[i] >= 'a' && output[i] <= 'z')))
	output[i] = '_';
      i++;
    }

  i = strlen (output);
  while (--i)
    if (output[i] < '.')
      break;

  if (i)
    output[i] = 0;

  return output;
}

void
create_package (void)
{
  FILE *fl;
  char *file;

  if (!
      (file =
       (char *) malloc (sizeof (char) * strlen (output_file.dir) + 15)))
    fatal ("Error: memory.");

  sprintf (file, "%s/Makefile.am", output_file.dir);

  if (!(fl = fopen (file, "w")))
    fatal ("Error write %s", file);

  fprintf (fl, "plugin_LTLIBRARIES = %s.la\n", output_file.no_ext);
  fprintf (fl, "plugindir = $(libdir)\n\n");
  fprintf (fl, "%s_la_SOURCES = %s\n", output_file.no_ext,
	   output_file.output);
  fprintf (fl, "%s_la_LDFLAGS = -module -avoid-version\n\n",
	   output_file.no_ext);

  fclose (fl);

  sprintf (file, "%s/configure.ac", output_file.dir);

  if (!(fl = fopen (file, "w")))
    fatal ("Error write %s", file);

  fprintf (fl, "AC_INIT(%s)\n", output_file.output);
  fprintf (fl, "AM_SANITY_CHECK\n");
  fprintf (fl, "AM_INIT_AUTOMAKE(%s,0.1)\n", output_file.no_ext);
  fprintf (fl, "AM_CONFIG_HEADER(config.h)\n");
  fprintf (fl, "AM_MAINTAINER_MODE\n\n");
  fprintf (fl, "AC_PROG_CC\n");
  fprintf (fl, "AC_ISC_POSIX\n");
  fprintf (fl, "AC_PROG_RANLIB\n");
  fprintf (fl, "AC_PROG_LIBTOOL\n");
  fprintf (fl, "AC_PROG_INSTALL\n");
  fprintf (fl, "AM_PROG_CC_STDC\n\n");
  fprintf (fl, "AC_HEADER_STDC\n");
  fprintf (fl, "AC_HEADER_DIRENT\n\n");
  fprintf (fl, "LDFLAGS=`" PREFIX "/ginfo-config --libs`\n");
  fprintf (fl, "CFLAGS=`" PREFIX "/ginfo-config --cflags`\n");
  fprintf (fl, "libdir=`" PREFIX "/ginfo-config --plugin_c_dir`\n");
  fprintf (fl, "AC_OUTPUT([Makefile])\n\n");

  fclose (fl);

  sprintf (file, "%s/autogen.sh", output_file.dir);

  if (!(fl = fopen (file, "w")))
    fatal ("Error write %s", file);

  fprintf (fl,
	   "#!/bin/bash\n\naclocal\nautoheader --warnings=none\nautoconf --warnings=none\nautomake --foreign -a\n\n");
  fclose (fl);

  chmod (file, 0750);

  free (file);
}

void
check_output_data (char *f)
{
  char *o = strdup (f);
  int i = strlen (f);

  if (!o)
    fatal ("Error: memory.");

  while (--i)
    {
      if (o[i] == '/')
	break;
    }

  if (i)
    {
      o[i] = 0;
      output_file.dir = o;
      o += i + 1;
      output_file.output = strdup (o);
    }
  else
    {
      output_file.dir = ".";
      output_file.output = strdup (f);
    }

  if (!output_file.output)
    fatal ("Error: memory.");

  i = strlen (output_file.output);
  while (--i)
    if (output_file.output[i] == '.')
      break;

  if (i && strlen (output_file.output) >= i + 3)
    {
      output_file.output[i + 1] = 'c';
      output_file.output[i + 2] = 0;
    }

  else
    {
      if (!
	  (o =
	   (char *) malloc (sizeof (char) *
			    (strlen (output_file.output) + 3))))
	fatal ("Error: memory.");
      sprintf (o, "%s.c", output_file.output);
      free (output_file.output);
      output_file.output = o;
    }

  if (!(output_file.no_ext = strdup (output_file.output)))
    fatal ("Error: memory.");

  i = strlen (output_file.no_ext);
  while (--i)
    if (output_file.no_ext[i] == '.')
      break;

  if (i)
    output_file.no_ext[i] = 0;

  while (i < strlen (output_file.no_ext))
    {
      if (!
	  ((output_file.no_ext[i] >= 'a' && output_file.no_ext[i] <= 'z')
	   || (output_file.no_ext[i] >= 'a' && output_file.no_ext[i] <= 'z')))
	output_file.no_ext[i] = '_';

      i++;
    }

  printf ("DIR: %s\n", output_file.dir);
  printf ("OUTPUT: %s\n", output_file.output);
  printf ("NO_EXT: %s\n", output_file.no_ext);
}

/* EOF */
